import unittest
from unittest import TestCase

import torchvision.models as models
import torchvision.transforms as transforms
from torch_dreams import Dreamer

from torch_dreams.auto_image_param import AutoImageParam
from torch_dreams.custom_image_param import CustomImageParam
from torch_dreams.masked_image_param import MaskedImageParam
from torch_dreams.transforms import random_resize   
from torch_dreams.model_bunch import ModelBunch
from torch_dreams.image_transforms import InverseTransform

import numpy as np
import torch
import os 

    
def make_custom_func(layer_number = 0, channel_number= 0): 
    def custom_func(layer_outputs):
        loss = layer_outputs[layer_number][:, channel_number].mean()
        return -loss
    return custom_func

class test(unittest.TestCase):

    def test_single_model(self):

        model = models.inception_v3(weights='DEFAULT')

        dreamy_boi = Dreamer(model = model, device= 'cpu', quiet= False)

        image_param = dreamy_boi.render(
            layers = [model.Mixed_6a],
            iters = 5
        )

        image_param.save(filename = 'test_single_model.jpg')

        self.assertTrue(os.path.exists('test_single_model.jpg'))
        self.assertTrue(isinstance(image_param, AutoImageParam), 'should be an instance of auto_image_param')
        self.assertTrue(isinstance(image_param.__array__(), np.ndarray))
        self.assertTrue(isinstance(image_param.to_hwc_tensor(), torch.Tensor), 'should be a torch.Tensor')
        self.assertTrue(isinstance(image_param.to_chw_tensor(), torch.Tensor), 'should be a torch.Tensor')
        os.remove('test_single_model.jpg')

    def test_custom_size(self):

        model = models.inception_v3(weights='DEFAULT')

        dreamy_boi = Dreamer(model = model, device= 'cpu', quiet= False)

        image_param = dreamy_boi.render(
            layers = [model.Mixed_6a],
            iters = 5,
            width = 255,
            height = 255
        )

        image_param.save(filename = 'test_custom_size.jpg')

        self.assertTrue(os.path.exists('test_custom_size.jpg'))
        self.assertTrue(isinstance(image_param, AutoImageParam), 'should be an instance of auto_image_param')
        self.assertTrue(isinstance(image_param.__array__(), np.ndarray))
        self.assertTrue(isinstance(image_param.to_hwc_tensor(), torch.Tensor), 'should be a torch.Tensor')
        self.assertTrue(isinstance(image_param.to_chw_tensor(), torch.Tensor), 'should be a torch.Tensor')
        os.remove('test_custom_size.jpg')

    def  test_single_model_custom_func(self):
        model = models.inception_v3(weights='DEFAULT')

        dreamy_boi = Dreamer(model = model, device= 'cpu', quiet= False)

        image_param = dreamy_boi.render(
            layers = [model.Mixed_6a],
            iters = 5,
            custom_func= make_custom_func(layer_number= 0, channel_number= 10)
        )

        image_param.save(filename = 'test_single_model_custom_func.jpg')
        
        self.assertTrue(os.path.exists('test_single_model_custom_func.jpg'))
        self.assertTrue(isinstance(image_param, AutoImageParam), 'should be an instance of auto_image_param')
        self.assertTrue(isinstance(image_param.__array__(), np.ndarray))
        self.assertTrue(isinstance(image_param.to_hwc_tensor(), torch.Tensor), 'should be a torch.Tensor')
        self.assertTrue(isinstance(image_param.to_chw_tensor(), torch.Tensor), 'should be a torch.Tensor')
        os.remove('test_single_model_custom_func.jpg')

    def test_multiple_models_custom_func(self):

        model1  = models.inception_v3(weights='DEFAULT').eval()
        model2  = models.resnet18(pretrained= True).eval() 

        bunch = ModelBunch(
            model_dict = {
                'inception': model1,
                'resnet':  model2
            }
        )

        layers_to_use = [
            bunch.model_dict['inception'].Mixed_6a,
            bunch.model_dict['resnet'].layer2[0].conv1
        ]

        dreamy_boi = Dreamer(model = bunch, quiet= False, device= 'cpu')

        def custom_func(layer_outputs):
            loss =  layer_outputs[1][:, 89].mean() + layer_outputs[0].mean()**2
            return -loss

        image_param = dreamy_boi.render(
            layers = layers_to_use,
            custom_func= custom_func,
            iters= 5
        )

        image_param.save(filename = 'test_multiple_models_custom_func.jpg')

        self.assertTrue(os.path.exists('test_multiple_models_custom_func.jpg'))
        self.assertTrue(isinstance(image_param, AutoImageParam), 'should be an instance of auto_image_param')
        self.assertTrue(isinstance(image_param.__array__(), np.ndarray))
        self.assertTrue(isinstance(image_param.to_hwc_tensor(), torch.Tensor), 'should be a torch.Tensor')
        self.assertTrue(isinstance(image_param.to_chw_tensor(), torch.Tensor), 'should be a torch.Tensor')

        os.remove('test_multiple_models_custom_func.jpg')

    def test_custom_image_param(self):

        model = models.inception_v3(weights='DEFAULT')

        dreamy_boi = Dreamer(model = model, device= 'cpu', quiet= False)
        param = CustomImageParam(image = 'images/sample_small.jpg', device= 'cpu')

        image_param = dreamy_boi.render(
            image_parameter= param,
            layers = [model.Mixed_6a],
            iters = 5,
            lr = 2e-4,
            grad_clip = 0.1,
            weight_decay= 1e-1
        )

        image_param.save(filename = 'test_custom_image_param.jpg')

        self.assertTrue(os.path.exists('test_custom_image_param.jpg'))
        self.assertTrue(isinstance(image_param, CustomImageParam), 'should be an instance of auto_image_param')
        self.assertTrue(isinstance(image_param.__array__(), np.ndarray))
        self.assertTrue(isinstance(image_param.to_hwc_tensor(), torch.Tensor), 'should be a torch.Tensor')
        self.assertTrue(isinstance(image_param.to_chw_tensor(), torch.Tensor), 'should be a torch.Tensor')
        os.remove('test_custom_image_param.jpg')

    def test_custom_image_param_set_param(self):
        """
        checks if custom_image_param.set_param correctly 
        loads the image without discrepancies with an absolute tolerance of 1e-5 element-wise
        """
        model = models.inception_v3(weights='DEFAULT')

        dreamy_boi = Dreamer(model = model, device= 'cpu', quiet= False)
        param = CustomImageParam(image = 'images/sample_small.jpg', device= 'cpu')

        image_param = dreamy_boi.render(
            image_parameter= param,
            layers = [model.Mixed_6a],
            iters = 5,
            lr = 2e-4,
            grad_clip = 0.1,
            weight_decay= 1e-1
        )

        image_tensor = image_param.to_nchw_tensor()

        image_param.set_param(tensor = image_tensor)
        # print(torch.abs((image_tensor - image_param.to_nchw_tensor())).mean())

        self.assertTrue(torch.allclose(image_tensor ,image_param.to_nchw_tensor(), atol = 1e-5))

    def test_MaskedImageParam(self):

        model = models.inception_v3(weights='DEFAULT')

        dreamy_boi = Dreamer(model = model, device= 'cpu', quiet= False)

        mask_tensor = torch.ones(1,3,512,512)
        mask_tensor[:,:,:256,:] = 0.

        param = MaskedImageParam(
            image = 'images/sample_small.jpg',
            mask_tensor = mask_tensor,
            device = 'cpu'
        )

        image_param = dreamy_boi.render(
            image_parameter= param,
            layers = [model.Mixed_6a],
            iters = 5,
            lr = 2e-4,
            grad_clip = 0.1,
            weight_decay= 1e-1
        )

        image_param.save(filename = 'test_MaskedImageParam.jpg')

        self.assertTrue(os.path.exists('test_MaskedImageParam.jpg'))
        self.assertTrue(isinstance(image_param, CustomImageParam), 'should be an instance of auto_image_param')
        self.assertTrue(isinstance(image_param.__array__(), np.ndarray))
        self.assertTrue(isinstance(image_param.to_hwc_tensor(), torch.Tensor), 'should be a torch.Tensor')
        self.assertTrue(isinstance(image_param.to_chw_tensor(), torch.Tensor), 'should be a torch.Tensor')
        os.remove('test_MaskedImageParam.jpg')     

    def test_caricature(self):

        model = models.resnet18(weights='DEFAULT')

        dreamy_boi = Dreamer(model, device = 'cpu')
        param = CustomImageParam(image = 'images/sample_small.jpg', device= 'cpu')

        image_tensor = param.to_nchw_tensor()

        param = dreamy_boi.caricature(
            input_tensor = image_tensor, 
            layers = [model.layer3],
            power= 1.0,
            iters = 5
        )

        self.assertTrue(isinstance(param, AutoImageParam), 'should be an auto_image_param')

    def test_static_caricature(self):

        model = models.resnet18(weights='DEFAULT')

        dreamy_boi = Dreamer(model, device = 'cpu')
        param = CustomImageParam(image = 'images/sample_small.jpg', device= 'cpu')

        image_tensor = param.to_nchw_tensor()

        param = dreamy_boi.caricature(
            input_tensor = image_tensor, 
            layers = [model.layer3],
            power= 1.0,
            iters = 5,
            static= True
        )

        self.assertTrue(isinstance(param, AutoImageParam), 'should be an auto_image_param')

    def test_custom_normalization(self):

        model = models.resnet18(weights='DEFAULT')

        dreamy_boi = Dreamer(model, device = 'cpu')

        t = transforms.Normalize(
                mean=[0.485, 0.456, 0.406],
                std=[0.229, 0.224, 0.225]
            )

        dreamy_boi.set_custom_normalization(normalization_transform = t)

        param = dreamy_boi.render(
            layers = [model.layer3],
            iters = 5
        )

        expected_transforms = transforms.Compose([

            transforms.RandomAffine(15, translate= (0,0)),
            random_resize(max_size_factor = 1.2, min_size_factor = 0.5),
            InverseTransform(
                old_mean = [0.485, 0.456, 0.406],
                old_std = [0.229, 0.224, 0.225],
                new_transforms = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])  
            )
        ])

        self.assertTrue(expected_transforms , dreamy_boi.transforms)

if __name__ == '__main__':

    unittest.main()



import torchvision.transforms as transforms
import torch.nn.functional as F
import torch.nn as nn


import warnings

warnings.filterwarnings("ignore")

transform_to_tensor = transforms.Compose([transforms.ToTensor()])

transform_and_rotate = transforms.Compose(
    [
        transforms.ToPILImage(),
        transforms.RandomRotation(
            degrees=5, expand=False, center=None, fill=None
        ),
        transforms.ToTensor(),
    ]
)

def unnormalize_image_tensor(img, mean, std):
    img = img * std.view(1,3,1,1) + mean.view(1,3,1,1)
    return img.clamp(0,1)


class InverseTransform(nn.Module):
    def __init__(self, old_mean, old_std, new_transforms):
        super().__init__()
        self.old_mean = old_mean
        self.old_std = old_std
        self.new_transform = new_transforms

    def forward(self, x):
        x = unnormalize_image_tensor(
            img = x,
            mean = self.old_mean,
            std = self.old_std
        )
        x = self.new_transform(x)
        return x

    def __repr__(self):
        return (
            "InverseTransform(\n        "
            + str(self.inverse_transform)
            + "\n        "
            + str(self.new_transform)
            + ")"
        )


def resize_4d_tensor_by_factor(x, height_factor, width_factor):
    res = F.interpolate(x, scale_factor=(height_factor, width_factor), mode="bilinear")
    return res


def resize_4d_tensor_by_size(x, height, width):
    res = F.interpolate(x, size=(height, width), mode="bilinear")
    return res



import torchvision.models as models
from torch_dreams import Dreamer

from torch_dreams.auto_image_param import AutoImageParam
from torch_dreams.batched_image_param import BatchedAutoImageParam
from torch_dreams.batched_objective import BatchedObjective
import numpy as np
import torch
import os

import pytest


def make_custom_func(layer_number=0, channel_number=0):
    def custom_func(layer_outputs):
        loss = layer_outputs[layer_number][:, channel_number].mean()
        return -loss

    return custom_func


@pytest.mark.parametrize("iters", [1, 2, 10, 20])
@pytest.mark.parametrize("batch_size", [1, 2, 5])
def test_batched_auto_image_param(iters, batch_size):

    model = models.inception_v3(weights='DEFAULT')
    dreamy_boi = Dreamer(model=model, device="cpu", quiet=False)

    image_param = BatchedAutoImageParam(batch_size=batch_size, device="cpu")

    result = dreamy_boi.render(
        layers=[model.Mixed_6a], iters=iters, image_parameter=image_param
    )

    for image_param in result.image_params:
        assert isinstance(
            image_param, AutoImageParam
        ), "should be an instance of auto_image_param"
        assert isinstance(image_param.__array__(), np.ndarray)
        assert isinstance(
            image_param.to_hwc_tensor(), torch.Tensor
        ), "should be a torch.Tensor"
        assert isinstance(
            image_param.to_chw_tensor(), torch.Tensor
        ), "should be a torch.Tensor"

    for i in range(batch_size):
        filename = f"test_single_model_iters_{iters}_batch_idx{i}.jpg"
        result[i].save(filename=filename)
        assert os.path.exists(filename)
        os.remove(filename)


@pytest.mark.parametrize("iters", [1, 2, 10, 20])
@pytest.mark.parametrize("batch_size", [1, 2, 5])
def test_batched_auto_image_param_with_custom_func(iters, batch_size):

    model = models.inception_v3(weights='DEFAULT')
    dreamy_boi = Dreamer(model=model, device="cpu", quiet=False)

    image_param = BatchedAutoImageParam(batch_size=batch_size, device="cpu")

    objective_functions = []
    for i in range(batch_size):
        objective_functions.append(make_custom_func(layer_number=0, channel_number=i))

    batched_objective = BatchedObjective(objectives=objective_functions)

    result = dreamy_boi.render(
        layers=[model.Mixed_6a],
        iters=iters,
        image_parameter=image_param,
        custom_func=batched_objective,
    )

    for image_param in result.image_params:
        assert isinstance(
            image_param, AutoImageParam
        ), "should be an instance of auto_image_param"
        assert isinstance(image_param.__array__(), np.ndarray)
        assert isinstance(
            image_param.to_hwc_tensor(), torch.Tensor
        ), "should be a torch.Tensor"
        assert isinstance(
            image_param.to_chw_tensor(), torch.Tensor
        ), "should be a torch.Tensor"



import torch.nn as nn


class ModelBunch:
    def __init__(self, model_dict: dict):
        self.model_dict = model_dict

    def forward(self, x):

        outs = []

        for model in list(self.model_dict.values()):
            output = model(x)
            outs.append(output)

        return outs

    def __call__(self, x):
        return self.forward(x)

    def eval(self):

        for model in list(self.model_dict.values()):
            model.eval()

    def to(self, device):
        for model in list(self.model_dict.values()):
            model.to(device)



import torch
import numpy as np

def default_func_mean(layer_outputs):
    """Default loss function for torch_dreams

    Args:
        layer_outputs (list): List of layers whose outputs are to be maximized. 

    Returns:
        [torch.tensor]: -loss
    """
    loss = 0.
    for out in layer_outputs:
        loss += out.mean()
    return -loss

class Hook():
    def __init__(self, module, backward=False):
        if backward==False:
            self.hook = module.register_forward_hook(self.hook_fn)
        else:
            self.hook = module.register_backward_hook(self.hook_fn)
    def hook_fn(self, module, input, output):
        self.input = input
        self.output = output
    def close(self):
        self.hook.remove()


import cv2
import torch
from .custom_image_param import CustomImageParam
from .transforms import imagenet_transform


class MaskedImageParam(CustomImageParam):
    def __init__(self, mask_tensor, image=None, device="cuda"):
        """Custom image param, but with a mask over the original image.

        The mask helps update only certain parts of the image
        and leave the rest untouched by training. Ideal for
        things like: "differentiable" backgrounds.

        Args:
            image (str or torch.tensor): "path/to/image.jpg" or NCHW tensor
            mask_tensor (torch.tensor): NCHW tensor whose values are clipped between 0,1
            device (str): 'cpu' or 'cuda'
        """

        self.width = mask_tensor.shape[-1]
        self.height = mask_tensor.shape[-2]

        if image is None:
            from .utils import init_image_param, fft_to_rgb, lucid_colorspace_to_rgb

            if self.width % 2 == 1:
                self.param = init_image_param(
                    height=self.height, width=self.width + 1, sd=0.01, device=device
                )
            else:
                self.param = init_image_param(
                    height=self.height, width=self.width, sd=0.01, device=device
                )
            img = fft_to_rgb(
                height=self.height,
                width=self.width,
                image_parameter=self.param,
                device=device,
            )
            img = lucid_colorspace_to_rgb(t=img, device=device)
            image = torch.sigmoid(img)

            # image = torch.ones(1,3,self.height,self.width)

        super().__init__(image=image, device=device)

        self.mask = mask_tensor.to(self.device)

        if isinstance(image, str):
            image = cv2.cvtColor(cv2.imread(image), cv2.COLOR_BGR2RGB) / 255.0
            image = torch.tensor(image).permute(-1, 0, 1).unsqueeze(0)

        self.original_nchw_image_tensor = image.to(device)

        assert (
            self.mask.shape[-2:] == self.to_nchw_tensor(device=self.device).shape[-2:]
        ), (
            "The height and width of the input image"
            + str(self.to_nchw_tensor(device=self.device).shape[-2:])
            + "and the mask"
            + str(self.mask.shape[-2:])
            + "do not match."
        )

    def to_chw_tensor(self, device="cpu"):
        t = self.forward(device=device).squeeze(0).clamp(
            0, 1
        ).detach() + self.original_nchw_image_tensor.to(device) * (
            1 - self.mask.to(device)
        )
        return t.squeeze(0)

    def forward(self, device):
        return self.normalize(self.postprocess(device=device), device=device).clamp(
            0, 1
        ) * self.mask.to(device)

    def update_mask(self, mask):
        """updates the mask to have a new value

        Warning: it also updates the original image

        Args:
            mask (torch.tensor): new mask to be used
        """

        self.original_nchw_image_tensor = self.to_chw_tensor(
            device=self.device
        ).unsqueeze(0)
        self.mask = mask.to(self.device)

    def to_hwc_tensor(self, device="cpu"):
        t = self.forward(device=device).squeeze(0).clamp(
            0, 1
        ).detach() + self.original_nchw_image_tensor.to(device) * (
            1 - self.mask.to(device)
        )
        return t.squeeze(0).permute(1, 2, 0)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()

"""
release checklist:
1. update version on `setup.py`
2. update `__version__` on `torch_dreams/__init__.py`
3. run tests with this command `pytest torch_dreams/tests/`
4. commit changes (`setup.py`, `torch_dreams/__init__.py`) and push
5. make release on PyPI. Run the following commands:
    5.1 `python3 setup.py sdist bdist_wheel`
    5.2 (optional) `python3 -m pip install --user --upgrade twine`
    5.3 `python3 -m twine upload dist/*`
6. make a new release on github with the latest version
"""

setuptools.setup(
    name="torch-dreams",
    version="4.0.0",
    author="Mayukh Deb",
    author_email="mayukhmainak2000@gmail.com",
    description="Making neural networks more interpretable, for research and art",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Mayukhdeb/torch-dreams",
    packages=setuptools.find_packages(),
    install_requires=required,
    python_requires=">=3.6",
    include_package_data=True,
    keywords=[
        "PyTorch",
        "machine learning",
        "neural networks",
        "convolutional neural networks",
        "feature visualization",
        "optimization",
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    test_suite="nose.collector",
    tests_require=["nose"],
)
